from __future__ import annotations

from enum import IntEnum
from typing import Literal


class CompressionType(IntEnum):
    """Compression algorithms for ZipFile"""

    STORED = 0
    """The numeric constant for an uncompressed archive member."""
    DEFLATED = 8
    """
    The numeric constant for the usual ZIP compression method. 
    This requires the [zlib](https://docs.python.org/3/library/zlib.html#module-zlib) module.
    """
    BZIP2 = 12
    """
    The numeric constant for the BZIP2 compression method. 
    This requires the [bz2](https://docs.python.org/3/library/bz2.html#module-bz2) module.
    """
    LZMA = 14
    """
    The numeric constant for the LZMA compression method. 
    This requires the [lzma](https://docs.python.org/3/library/lzma.html#module-lzma) module.
    """

    @classmethod
    def get(
        cls,
        key: str | int | None = None,
        default: Literal["stored", "deflated", "bzip2", "lzma"] = "stored",
    ) -> CompressionType:
        """
        Get the `CompressionType` by its name or number.
        Return the default if the key is missing or invalid.

        Parameters
        ----------
        key : str | int, optional
            The key to retrieve.

        Returns
        -------
        CompressionType
            The `CompressionType` corresponding to the key.
        """
        try:
            match key:
                case str():
                    return cls[key.upper()]

                case int():
                    return cls(key)

                case _:
                    return cls[default.upper()]
        except (KeyError, ValueError):
            return cls[default.upper()]



from __future__ import annotations

import re
from pathlib import Path
from tarfile import is_tarfile
from types import TracebackType
from typing import Any, overload
from zipfile import is_zipfile

from py7zr import is_7zfile
from pydantic import validate_call
from rarfile import is_rarfile, is_rarfile_sfx
from typing_extensions import Generator, Self

from archivefile._adapters._base import BaseArchiveAdapter
from archivefile._adapters._rar import RarFileAdapter
from archivefile._adapters._sevenzip import SevenZipFileAdapter
from archivefile._adapters._tar import TarFileAdapter
from archivefile._adapters._zip import ZipFileAdapter
from archivefile._enums import CompressionType
from archivefile._models import ArchiveMember
from archivefile._types import (
    CollectionOf,
    CompressionLevel,
    ErrorHandler,
    OpenArchiveMode,
    SortBy,
    StrPath,
    TableStyle,
    TreeStyle,
)
from archivefile._utils import realpath


class ArchiveFile(BaseArchiveAdapter):
    # fmt: off
    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...
    # fmt: on

    @validate_call
    def __init__(
        self,
        file: StrPath,
        mode: OpenArchiveMode | str = "r",
        *,
        password: str | None = None,
        compression_type: CompressionType | None = None,
        compression_level: CompressionLevel | int | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Open an archive file.

        Parameters
        ----------
        file : StrPath
            Path to the archive file.
        mode : OpenArchiveMode, optional
            Mode for opening the archive file.
        password : str, optional
            Password for encrypted archive files.
        compression_type : CompressionType, optional
            The compression method for writing zip files.
        compression_level : CompressionLevel, optional
            The compression level for writing zip files.
        kwargs : Any
            Keyword arugments to pass to the underlying library.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            Raised if the archive format is unsupported

        Notes
        -----
        The `compression_type` and `compression_level` parameters are only applicable when creating
        zip files and do not affect reading zip files or other archive formats.

        References
        ----------
        ArchiveFile currently supports the following:

        - [`ZipFile`][zipfile.ZipFile]
        - [`TarFile`][tarfile.TarFile.open]
        - [`RarFile`][rarfile.RarFile]
        - [`SevenZipFile`][py7zr.SevenZipFile]
        """
        self._file = realpath(file)
        self._mode = mode
        self._password = password
        self._kwargs = kwargs
        self._compression_type = compression_type
        self._compression_level = compression_level
        self._initialize_adapter()

    def _initialize_adapter(self) -> None:
        if not self._file.exists():
            if not self._mode.startswith("r"):
                filename = self._file.name.lower()

                if re.search(r"\.(zip|cbz)$", filename):
                    adapter = ZipFileAdapter
                elif re.search(r"\.((tar(\.(bz2|gz|xz))?)|(cbt))$", filename):
                    adapter = TarFileAdapter  # type: ignore
                elif re.search(r"\.(7z|cb7)$", filename):
                    adapter = SevenZipFileAdapter  # type: ignore
                elif re.search(r"\.(rar|cbr)$", filename):
                    adapter = RarFileAdapter  # type: ignore
                else:
                    raise NotImplementedError(f"Unsupported archive format: {self._file}")
            else:
                raise FileNotFoundError(self._file)

        elif is_zipfile(self._file):
            adapter = ZipFileAdapter

        elif is_tarfile(self._file):
            adapter = TarFileAdapter  # type: ignore

        elif is_7zfile(self._file):
            adapter = SevenZipFileAdapter  # type: ignore

        elif is_rarfile(self._file) or is_rarfile_sfx(self._file):
            adapter = RarFileAdapter  # type: ignore

        else:
            raise NotImplementedError(f"Unsupported archive format: {self._file}")

        self._adapter = adapter(
            self._file,
            self._mode,
            password=self._password,
            compression_type=self._compression_type,
            compression_level=self._compression_level,
            **self._kwargs,
        )

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        self._adapter.close()

    @property
    def file(self) -> Path:
        """Path to the archive file."""
        return self._adapter.file

    @property
    def mode(self) -> OpenArchiveMode:
        """Mode in which the archive file was opened."""
        return self._adapter.mode

    @property
    def password(self) -> str | None:
        """Archive password."""
        return self._adapter.password

    @property
    def compression_type(self) -> CompressionType | None:
        """Compression type used for writing."""
        return self._adapter.compression_type

    @property
    def compression_level(self) -> CompressionLevel | None:
        """Compression level used for writing."""
        return self._adapter.compression_level  # type: ignore

    @property
    def adapter(self) -> str:
        """Name of the underlying adapter class, useful for debugging."""
        return self._adapter.adapter

    @validate_call
    def get_member(self, member: StrPath) -> ArchiveMember:
        """
        Retrieve an ArchiveMember object by it's name.

        Parameters
        ----------
        member : StrPath
            Name of the member.

        Returns
        -------
        ArchiveMember
            Represents a member of the archive.

        Raises
        ------
        KeyError
            Raised if the member is not found in the archive.

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.tar") as archive:
            archive.get_member("README.md")
            # ArchiveMember(name='README.md', size=3799, compressed_size=3799, datetime=datetime.datetime(2024, 4, 10, 20, 10, 57, tzinfo=datetime.timezone.utc), checksum=5251, is_dir=False, is_file=True)
        ```
        """
        return self._adapter.get_member(member)

    @validate_call
    def get_members(self) -> Generator[ArchiveMember]:
        """
        Retrieve all members of the archive as a generator of `ArchiveMember` objects.

        Yields
        -------
        ArchiveMember
            Each member of the archive as an `ArchiveMember` object.

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.tar") as archive:
            for member in archive.get_members():
                print(member.name)
                # project/pyproject.toml
                # project/src
        ```
        """
        return self._adapter.get_members()

    @validate_call
    def get_names(self) -> tuple[str, ...]:
        """
        Retrieve all members of the archive as a tuple of strings.

        Returns
        -------
        tuple[str, ...]
            Members of the archive as a tuple of strings.

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.tar") as archive:
            archive.get_names()
            # (
            #     "project/pyproject.toml",
            #     "project/src",
            # )
        ```
        """
        return self._adapter.get_names()

    def print_tree(
        self,
        *,
        max_depth: int = 0,
        style: TreeStyle = "const",
    ) -> None:
        """
        Print the contents of the archive as a tree.

        Parameters
        ----------
        max_depth : int, optional
            Maximum depth to print.
        style : TreeStyle, optional
            The style of the tree.

        Returns
        -------
        None

        Notes
        -----
        The [`bigtree`](https://pypi.org/p/bigtree/) dependency is required to use this method.

        You can install it via either of these commands:

        - `pip install archivefile[bigtree]`
        - `pip install archivefile[all]`

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.tar.gz") as archive:
            archive.print_tree()
            # source.tar.gz
            # └── hello-world
            #     ├── pyproject.toml
            #     ├── README.md
            #     ├── src
            #     │   └── hello_world
            #     │       └── __init__.py
            #     └── tests
            #         └── __init__.py
        ```
        """
        self._adapter.print_tree(max_depth=max_depth, style=style)

    @validate_call
    def print_table(
        self,
        *,
        title: str | None = None,
        style: TableStyle = "markdown",
        sort_by: SortBy = "name",
        descending: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Print the contents of the archive as a table.

        Parameters
        ----------
        title : str, optional
            Title of the table. Defaults to archive file name.
        style : TableStyle, optional
            The style of the table.
        sort_by : SortBy, optional
            Key used to sort the table.
        descending : bool, optional
            If True, sorting will be in descending order.
        kwargs : Any
            Additional keyword arguments to be passed to the [`Table`][rich.table.Table] constructor.

        Returns
        -------
        None

        Notes
        -----
        The [`rich`](https://pypi.org/p/rich/) dependency is required to use this method.

        You can install it via either of these commands:

        - `pip install archivefile[rich]`
        - `pip install archivefile[all]`

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.zip") as archive:
            archive.print_table()
            #                                                     source.zip
            #
            # | Name                                    | Date modified             | Type   | Size | Compressed Size |
            # |-----------------------------------------|---------------------------|--------|------|-----------------|
            # | hello-world/                            | 2024-05-02T09:41:24+00:00 | Folder | 0B   | 0B              |
            # | hello-world/README.md                   | 2024-05-02T09:41:24+00:00 | File   | 0B   | 0B              |
            # | hello-world/pyproject.toml              | 2024-05-02T09:41:24+00:00 | File   | 363B | 241B            |
            # | hello-world/src/                        | 2024-05-02T09:41:24+00:00 | Folder | 0B   | 0B              |
            # | hello-world/src/hello_world/            | 2024-05-02T09:41:24+00:00 | Folder | 0B   | 0B              |
            # | hello-world/src/hello_world/__init__.py | 2024-05-02T09:41:24+00:00 | File   | 0B   | 0B              |
            # | hello-world/tests/                      | 2024-05-02T09:41:24+00:00 | Folder | 0B   | 0B              |
            # | hello-world/tests/__init__.py           | 2024-05-02T09:41:24+00:00 | File   | 0B   | 0B              |
        ```
        """
        self._adapter.print_table(title=title, style=style, sort_by=sort_by, descending=descending, **kwargs)

    @validate_call
    def extract(self, member: StrPath | ArchiveMember, *, destination: StrPath = Path.cwd()) -> Path:
        """
        Extract a member of the archive.

        Parameters
        ----------
        member : StrPath, ArchiveMember
            Name of the member or an ArchiveMember object.
        destination : StrPath
            The path to the directory where the member will be extracted.
            If not specified, the current working directory is used as the default destination.

        Returns
        -------
        Path
            The path to the extracted file.

        Raises
        ------
        KeyError
            Raised if the member is not found in the archive.

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.zip") as archive:
            file = archive.extract("hello-world/pyproject.toml")
            print(file.read_text())
            # [tool.poetry]
            # name = "hello-world"
            # version = "0.1.0"
            # description = ""
            # readme = "README.md"
            # packages = [{include = "hello_world", from = "src"}]
        ```
        """
        return self._adapter.extract(member, destination=destination)

    @validate_call
    def extractall(
        self, *, destination: StrPath = Path.cwd(), members: CollectionOf[StrPath | ArchiveMember] | None = None
    ) -> Path:
        """
        Extract all the members of the archive to the destination directory.

        Parameters
        ----------
        destination : StrPath
            The path to the directory where the members will be extracted.
            If not specified, the current working directory is used as the default destination.
        members : CollectionOf[StrPath | ArchiveMember], optional
            Collection of member names or ArchiveMember objects to extract.
            Default is `None` which will extract all members.

        Returns
        -------
        Path
            The path to the destination directory.

        Raises
        ------
        KeyError
            Raised if any member in members was not found in the archive.

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.zip") as archive:
            outdir = archive.extractall()

        for file in outdir.rglob("*"):
            print(file)
            # /source/hello-world
            # /source/hello-world/pyproject.toml
            # /source/hello-world/README.md
            # /source/hello-world/src
            # /source/hello-world/tests
            # /source/hello-world/src/hello_world
            # /source/hello-world/src/hello_world/__init__.py
            # /source/hello-world/tests/__init__.py
        ```
        """
        return self._adapter.extractall(destination=destination, members=members)

    @validate_call
    def read_bytes(self, member: StrPath | ArchiveMember) -> bytes:
        """
        Read the member in bytes mode.

        Parameters
        ----------
        member : StrPath, ArchiveMember
            Name of the member or an ArchiveMember object.

        Returns
        -------
        bytes
            The contents of the file as bytes.

        Raises
        ------
        KeyError
            Raised if the member is not found in the archive.

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.zip") as archive:
            data = archive.read_bytes("hello-world/pyproject.toml")
            print(data)
            # b'[tool.poetry]\\r\\nname = "hello-world"\\r\\nversion = "0.1.0"\\r\\ndescription = ""\\r\\nreadme = "README.md"\\r\\npackages = [{include = "hello_world", from = "src"}]\\r\\n'
        ```
        """
        return self._adapter.read_bytes(member)

    @validate_call
    def read_text(
        self,
        member: StrPath | ArchiveMember,
        *,
        encoding: str = "utf-8",
        errors: ErrorHandler = "strict",
    ) -> str:
        """
        Read the member in text mode.

        Parameters
        ----------
        member : StrPath, ArchiveMember
            Name of the member or an ArchiveMember object.
        encoding : str, optional
            Encoding used to read the file. Default is `utf-8`.
        errors : ErrorHandler, optional
            String that specifies how encoding and decoding errors are to be handled.

        Returns
        -------
        str
            The contents of the file as a string.

        Raises
        ------
        KeyError
            Raised if the member is not found in the archive.

        References
        ----------
        - [Standard Encodings](https://docs.python.org/3/library/codecs.html#standard-encodings)
        - [Error Handlers](https://docs.python.org/3/library/codecs.html#error-handlers)

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.zip") as archive:
            text = archive.read_text("hello-world/pyproject.toml")
            print(text)
            # [tool.poetry]
            # name = "hello-world"
            # version = "0.1.0"
            # description = ""
            # readme = "README.md"
            # packages = [{include = "hello_world", from = "src"}]
        ```
        """
        return self._adapter.read_text(member, encoding=encoding, errors=errors)

    @validate_call
    def write(
        self,
        file: StrPath,
        *,
        arcname: StrPath | None = None,
    ) -> None:
        """
        Write a single file to the archive.

        Parameters
        ----------
        file : StrPath
            Path of the file.
        arcname : StrPath, optional
            Name which the file will have in the archive.
            Default is the basename of the file.

        Returns
        -------
        None

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("example.zip", "w") as archive:
            archive.write("/root/some/path/to/foo.txt")
            archive.write("bar.txt", arcname="baz.txt")
            archive.write("recipe/spam.txt", arcname="ingredients/spam.txt")
            archive.write("recipe/eggs.txt", arcname="ingredients/eggs.txt")
            archive.print_tree()
            # example.zip
            # ├── foo.txt
            # ├── baz.txt
            # └── ingredients
            #     ├── spam.txt
            #     └── eggs.txt
        ```
        """
        self._adapter.write(file, arcname=arcname)

    @validate_call
    def write_text(
        self,
        data: str,
        *,
        arcname: StrPath,
    ) -> None:
        """
        Write the string `data` to a file within the archive named `arcname`.

        Parameters
        ----------
        data : str
            The text data to write to the archive.
        arcname : StrPath, optional
            The name which the file will have in the archive.

        Returns
        -------
        None

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("newarchive.zip", "w") as archive:
            archive.write_text("spam and eggs", arcname="recipe.txt")
            members = archive.get_names()
            print(members)
            # ('recipe.txt',)
            text = archive.read_text("recipe.txt")
            print(text)
            # 'spam and eggs'
        ```
        """
        self._adapter.write_text(data, arcname=arcname)

    @validate_call
    def write_bytes(
        self,
        data: bytes,
        *,
        arcname: StrPath,
    ) -> None:
        """
        Write the bytes `data` to a file within the archive named `arcname`.

        Parameters
        ----------
        data: str
            The bytes data to write to the archive.
        arcname : StrPath, optional
            The name which the file will have in the archive.

        Returns
        -------
        None

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("skynet.7z", "w") as archive:
            archive.write_bytes(b"010010100101", arcname="terminator.py")
            members = archive.get_names()
            print(members)
            # ('terminator.py',)
            data = archive.read_bytes("terminator.py")
            print(data)
            # b"010010100101"
        ```
        """
        self._adapter.write_bytes(data, arcname=arcname)

    @validate_call
    def writeall(
        self,
        dir: StrPath,
        *,
        root: StrPath | None = None,
        glob: str = "*",
        recursive: bool = True,
    ) -> None:
        """
        Write a directory to the archive.

        Parameters
        ----------
        dir : StrPath
            Path of the directory.
        root : StrPath
            Directory that will be the root directory of the archive, all paths in the archive will be relative to it.
            This must be relative to given directory path. Default is the parent of the given directory.
        glob : str, optional
            Only write files that match this glob pattern to the archive.
        recursive : bool, optional
            Recursively write all the files in the given directory. Default is True.

        Returns
        -------
        None

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        with ArchiveFile("source.tar.gz", "w:gz") as archive:
            archive.writeall(dir="hello-world/")
            archive.print_tree()
            # source.tar.gz
            # └── hello-world
            #     ├── pyproject.toml
            #     ├── README.md
            #     ├── src
            #     │   └── hello_world
            #     │       └── __init__.py
            #     └── tests
            #         └── __init__.py
        ```
        """

        self._adapter.writeall(
            dir,
            root=root,
            glob=glob,
            recursive=recursive,
        )

    def close(self) -> None:
        """
        Close the archive file.

        Returns
        -------
        None

        Examples
        --------
        ```py
        from archivefile import ArchiveFile

        archive = ArchiveFile("skynet.zip", "w")
        archive.write_bytes(b"01010101001", arcname="terminator.py")
        archive.close()
        ```
        """
        self._adapter.close()

    def __repr__(self) -> str:
        password = '"********"' if self.password else None
        return f'{self.__class__.__name__}("{self.file}", "{self.mode}", password={password})'



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, overload

if TYPE_CHECKING:
    from types import TracebackType

    from archivefile._enums import CompressionType
    from archivefile._models import ArchiveMember
    from archivefile._types import (
        CollectionOf,
        CompressionLevel,
        ErrorHandler,
        OpenArchiveMode,
        SortBy,
        StrPath,
        TableStyle,
        TreeStyle,
    )
    from typing_extensions import Generator, Self


class BaseArchiveAdapter(Protocol):
    """
    A base protocol that can be inherited from to implement more adapters.
    Refer to `src/archivefile/_core.py` for documentation of every method and property.
    """
    # fmt: off
    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...
    # fmt: on

    def __init__(
        self,
        file: StrPath,
        mode: OpenArchiveMode | str = "r",
        *,
        password: str | None = None,
        compression_type: CompressionType | None = None,
        compression_level: CompressionLevel | int | None = None,
        **kwargs: Any,
    ) -> None: ...

    def __enter__(self) -> Self: ...

    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None: ...

    @property
    def file(self) -> Path: ...

    @property
    def mode(self) -> OpenArchiveMode: ...

    @property
    def password(self) -> str | None: ...

    @property
    def compression_type(self) -> CompressionType | None: ...

    @property
    def compression_level(self) -> CompressionLevel | None: ...

    @property
    def adapter(self) -> str: ...

    def get_member(self, member: StrPath) -> ArchiveMember: ...

    def get_members(self) -> Generator[ArchiveMember]: ...

    def get_names(self) -> tuple[str, ...]: ...

    def print_tree(
        self,
        *,
        max_depth: int = 0,
        style: TreeStyle = "const",
    ) -> None: ...

    def print_table(
        self,
        *,
        title: str | None = None,
        style: TableStyle = "markdown",
        sort_by: SortBy = "name",
        descending: bool = False,
        **kwargs: Any,
    ) -> None: ...

    def extract(self, member: StrPath | ArchiveMember, *, destination: StrPath = Path.cwd()) -> Path: ...

    def extractall(
        self, *, destination: StrPath = Path.cwd(), members: CollectionOf[StrPath | ArchiveMember] | None = None
    ) -> Path: ...

    def read_bytes(self, member: StrPath | ArchiveMember) -> bytes: ...

    def read_text(
        self,
        member: StrPath | ArchiveMember,
        *,
        encoding: str = "utf-8",
        errors: ErrorHandler = "strict",
    ) -> str: ...

    def write(
        self,
        file: StrPath,
        *,
        arcname: StrPath | None = None,
    ) -> None: ...

    def write_text(
        self,
        data: str,
        *,
        arcname: StrPath,
    ) -> None: ...

    def write_bytes(
        self,
        data: bytes,
        *,
        arcname: StrPath,
    ) -> None: ...

    def writeall(
        self,
        dir: StrPath,
        *,
        root: StrPath | None = None,
        glob: str = "*",
        recursive: bool = True,
    ) -> None: ...

    def close(self) -> None: ...

    def __repr__(self) -> str:
        password = '"********"' if self.password else None
        return f'{self.__class__.__name__}("{self.file}", "{self.mode}", password={password})'



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, overload

from archivefile._adapters._base import BaseArchiveAdapter
from archivefile._models import ArchiveMember
from archivefile._utils import get_member_name, realpath
from py7zr import SevenZipFile

if TYPE_CHECKING:
    from types import TracebackType

    from archivefile._enums import CompressionType
    from archivefile._types import (
        CollectionOf,
        CompressionLevel,
        ErrorHandler,
        OpenArchiveMode,
        SortBy,
        StrPath,
        TableStyle,
        TreeStyle,
    )
    from typing_extensions import Generator, Self


class SevenZipFileAdapter(BaseArchiveAdapter):
    # fmt: off
    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...
    # fmt: on

    def __init__(
        self,
        file: StrPath,
        mode: OpenArchiveMode | str = "r",
        *,
        password: str | None = None,
        compression_type: CompressionType | None = None,
        compression_level: CompressionLevel | int | None = None,
        **kwargs: Any,
    ) -> None:
        self._file = realpath(file)
        self._mode = mode[0]
        self._password = password

        # Bit of a hack to support 'x' and 'a' modes properly
        if self._mode == "x":
            if self._file.exists():
                # SevenZipFile doesn't support mode='x'
                # open for exclusive creation, failing if the file already exists
                # see: https://github.com/miurahr/py7zr/issues/587
                raise FileExistsError(self._file)
            else:
                # 'x' and 'w' are equivalent if the file doesn't exist
                self._mode = "w"

        if self._mode == "a" and not self._file.exists():
            # SevenZipFile only supports mode='a' on existing files
            # Since 'a' is equivalent to 'w' in this case (i.e, when archive file doesn't already exist)
            # We can just set it to 'w'
            self._mode = "w"

        self._sevenzipfile = SevenZipFile(self._file, mode=self._mode, password=self._password, **kwargs)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        self._sevenzipfile.close()  # type: ignore

    @property
    def file(self) -> Path:
        return self._file

    @property
    def mode(self) -> OpenArchiveMode:
        return self._mode  # type: ignore

    @property
    def password(self) -> str | None:
        return self._password

    @property
    def compression_type(self) -> CompressionType | None:
        # SevenZipFile doesn't support this, so this will always be None.
        # Only ZipFile supports this
        return None

    @property
    def compression_level(self) -> CompressionLevel | None:
        # SevenZipFile doesn't support this, so this will always be None.
        # Only ZipFile supports this
        return None

    @property
    def adapter(self) -> str:
        return self.__class__.__name__

    def get_member(self, member: StrPath) -> ArchiveMember:
        # Unlike the rest, SevenZip member directories do not end with `/`, so we need to strip it out
        # i.e, `spam/eggs/` in a ZipFile is equivalent to `spam/eggs` in SevenZipFile
        name = get_member_name(member).removesuffix("/")

        # SevenZipFile doesn't have an equivalent for `get_member` like the rest, so we hand craft it instead
        # https://more-itertools.readthedocs.io/en/stable/_modules/more_itertools/recipes.html#first_true
        sevenzipinfo = next(filter(lambda mem: mem.filename == name, self._sevenzipfile.list()), None)

        # ZipFile and TarFile raise KeyError
        # So for consistency (and because I like KeyError over None), we'll also raise KeyError here
        if sevenzipinfo is None:
            raise KeyError(f"{name} not found in {self._file}")

        return ArchiveMember(
            name=sevenzipinfo.filename,
            size=sevenzipinfo.uncompressed,
            # Sometimes sevenzip can return 0 for compressed size when there's no compression
            # in that case we simply return the uncompressed size instead.
            compressed_size=sevenzipinfo.compressed or sevenzipinfo.uncompressed,
            datetime=sevenzipinfo.creationtime,
            checksum=sevenzipinfo.crc32,
            is_dir=sevenzipinfo.is_directory,
            is_file=not sevenzipinfo.is_directory,
        )

    def get_members(self) -> Generator[ArchiveMember]:
        return (
            ArchiveMember(
                name=sevenzipinfo.filename,
                size=sevenzipinfo.uncompressed,
                # Sometimes sevenzip can return 0 for compressed size when there's no compression
                # in that case we simply return the uncompressed size instead.
                compressed_size=sevenzipinfo.compressed or sevenzipinfo.uncompressed,
                datetime=sevenzipinfo.creationtime,
                checksum=sevenzipinfo.crc32,
                is_dir=sevenzipinfo.is_directory,
                is_file=not sevenzipinfo.is_directory,
            )
            for sevenzipinfo in self._sevenzipfile.list()
        )

    def get_names(self) -> tuple[str, ...]:
        return tuple(self._sevenzipfile.getnames())

    def print_tree(
        self,
        *,
        max_depth: int = 0,
        style: TreeStyle = "const",
    ) -> None:
        try:
            from bigtree.tree.construct import list_to_tree
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("The 'print_tree()' method requires the 'bigtree' dependency.")

        paths = [f"{self.file.name}/{member}" for member in self.get_names()]
        tree = list_to_tree(paths)  # type: ignore
        tree.show(max_depth=max_depth, style=style)

    def print_table(
        self,
        *,
        title: str | None = None,
        style: TableStyle = "markdown",
        sort_by: SortBy = "name",
        descending: bool = False,
        **kwargs: Any,
    ) -> None:
        try:
            from rich import box as RichBox
            from rich import print as richprint
            from rich.table import Table as RichTable
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("The 'print_table()' method requires the 'rich' dependency.")

        if title is None:
            title = self.file.name

        members = self.get_members()
        table = RichTable(title=title, box=getattr(RichBox, style.upper()), **kwargs)
        table.add_column("Name", no_wrap=True)
        table.add_column("Date modified", no_wrap=True)
        table.add_column("Type", no_wrap=True)
        table.add_column("Size", no_wrap=True)
        table.add_column("Compressed Size", no_wrap=True)
        for member in sorted(members, key=lambda member: getattr(member, sort_by), reverse=descending):
            table.add_row(
                member.name,
                member.datetime.isoformat(),
                "File" if member.is_file else "Folder",
                member.size.human_readable(),
                member.compressed_size.human_readable(),
            )
        richprint(table)

    def extract(self, member: StrPath | ArchiveMember, *, destination: StrPath = Path.cwd()) -> Path:
        destination = realpath(destination)
        destination.mkdir(parents=True, exist_ok=True)

        # Unlike the rest, SevenZip member directories do not end with `/`, so we need to strip it out
        # i.e, `spam/eggs/` in a ZipFile is equivalent to `spam/eggs` in SevenZipFile
        name = get_member_name(member).removesuffix("/")

        if name in self.get_names():
            self._sevenzipfile.extract(path=destination, targets=[name], recursive=True)
        else:
            # ZipFile and TarFile raise KeyError but SevenZipFile does nothing
            # So for consistency's sake, we'll also raise KeyError here
            raise KeyError(f"{name} not found in {self._file}")

        self._sevenzipfile.reset()
        return destination / name

    def extractall(
        self, *, destination: StrPath = Path.cwd(), members: CollectionOf[StrPath | ArchiveMember] | None = None
    ) -> Path:
        destination = realpath(destination)
        destination.mkdir(parents=True, exist_ok=True)

        names: set[str] = set()
        if members:
            all_members = self._sevenzipfile.getnames()
            for member in members:
                # Unlike the rest, SevenZip member directories do not end with `/`, so we need to strip it out
                # i.e, `spam/eggs/` in a ZipFile is equivalent to `spam/eggs` in SevenZipFile
                name = get_member_name(member).removesuffix("/")
                if name in all_members:
                    names.add(name)
                else:
                    raise KeyError(f"{name} not found in {self._file}")

        if names:
            self._sevenzipfile.extract(path=destination, targets=names, recursive=True)
        else:
            self._sevenzipfile.extractall(path=destination)

        self._sevenzipfile.reset()
        return destination

    def read_bytes(self, member: StrPath | ArchiveMember) -> bytes:
        # Unlike the rest, SevenZip member directories do not end with `/`, so we need to strip it out
        # i.e, `spam/eggs/` in a ZipFile is equivalent to `spam/eggs` in SevenZipFile
        name = get_member_name(member).removesuffix("/")

        if name not in self._sevenzipfile.getnames():
            raise KeyError(f"{name} not found in {self._file}")

        datadict = self._sevenzipfile.read(targets=[name])

        if fileobj := datadict.get(name):
            data = fileobj.read()
        else:
            data = b""

        self._sevenzipfile.reset()
        return data

    def read_text(
        self,
        member: StrPath | ArchiveMember,
        *,
        encoding: str = "utf-8",
        errors: ErrorHandler = "strict",
    ) -> str:
        return self.read_bytes(member).decode(encoding, errors)

    def write(
        self,
        file: StrPath,
        *,
        arcname: StrPath | None = None,
    ) -> None:
        file = realpath(file)

        match arcname:
            case None:
                arcname = file.name
            case Path():
                arcname = arcname.relative_to(arcname.anchor).as_posix()

        if not file.is_file():
            raise ValueError(f"The specified file '{file}' either does not exist or is not a regular file!")

        self._sevenzipfile.write(file, arcname=arcname)

    def write_text(
        self,
        data: str,
        *,
        arcname: StrPath,
    ) -> None:
        self._sevenzipfile.writestr(data=data, arcname=get_member_name(arcname))

    def write_bytes(
        self,
        data: bytes,
        *,
        arcname: StrPath,
    ) -> None:
        self._sevenzipfile.writestr(data=data, arcname=get_member_name(arcname))

    def writeall(
        self,
        dir: StrPath,
        *,
        root: StrPath | None = None,
        glob: str = "*",
        recursive: bool = True,
    ) -> None:
        dir = realpath(dir)

        if not dir.is_dir():
            raise ValueError(f"The specified file '{dir}' either does not exist or is not a regular directory!")

        if root is None:
            root = dir.parent
        else:
            root = realpath(root)

        if not dir.is_relative_to(root):
            raise ValueError(f"{dir} must be relative to {root}")

        files = dir.rglob(glob) if recursive else dir.glob(glob)

        for file in files:
            arcname = file.relative_to(root)
            self.write(file, arcname=arcname)

    def close(self) -> None:
        self._sevenzipfile.close()  # type: ignore

    def __repr__(self) -> str:
        password = '"********"' if self.password else None
        return f'{self.__class__.__name__}("{self.file}", "{self.mode}", password={password})'



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, overload
from zipfile import ZipFile

from archivefile._adapters._base import BaseArchiveAdapter
from archivefile._enums import CompressionType
from archivefile._models import ArchiveMember
from archivefile._utils import clamp_compression_level, get_member_name, realpath

if TYPE_CHECKING:
    from types import TracebackType

    from archivefile._types import (
        CollectionOf,
        CompressionLevel,
        ErrorHandler,
        OpenArchiveMode,
        SortBy,
        StrPath,
        TableStyle,
        TreeStyle,
    )
    from typing_extensions import Generator, Self


class ZipFileAdapter(BaseArchiveAdapter):
    # fmt: off
    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...
    # fmt: on

    def __init__(
        self,
        file: StrPath,
        mode: OpenArchiveMode | str = "r",
        *,
        password: str | None = None,
        compression_type: CompressionType | None = None,
        compression_level: CompressionLevel | int | None = None,
        **kwargs: Any,
    ) -> None:
        self._file = realpath(file)
        self._mode = mode[0]
        self._password = password
        self._pwd = password.encode() if password else None

        self._compression_type = CompressionType.get(compression_type)
        self._compression_level = clamp_compression_level(compression_level) if compression_level is not None else None

        if self._compression_type is CompressionType.BZIP2:
            # BZIP2 only supports 1-9
            if self._compression_level == 0:
                self._compression_level = 1

        self._zipfile = ZipFile(
            self._file,
            mode=self._mode,
            compression=self._compression_type,
            compresslevel=self._compression_level,
            **kwargs,
        )  # type: ignore

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        self._zipfile.close()

    @property
    def file(self) -> Path:
        return self._file

    @property
    def mode(self) -> OpenArchiveMode:
        return self._mode  # type: ignore

    @property
    def password(self) -> str | None:
        return self._password

    @property
    def compression_type(self) -> CompressionType | None:
        return self._compression_type

    @property
    def compression_level(self) -> CompressionLevel | None:
        return self._compression_level  # type: ignore

    @property
    def adapter(self) -> str:
        return self.__class__.__name__

    def get_member(self, member: StrPath) -> ArchiveMember:
        name = get_member_name(member)
        zipinfo = self._zipfile.getinfo(name)

        return ArchiveMember(
            name=zipinfo.filename,
            size=zipinfo.file_size,
            compressed_size=zipinfo.compress_size,
            datetime=datetime(*zipinfo.date_time),
            checksum=zipinfo.CRC,
            is_dir=zipinfo.is_dir(),
            is_file=not zipinfo.is_dir(),
        )

    def get_members(self) -> Generator[ArchiveMember]:
        return (
            ArchiveMember(
                name=zipinfo.filename,
                size=zipinfo.file_size,
                compressed_size=zipinfo.compress_size,
                datetime=datetime(*zipinfo.date_time),
                checksum=zipinfo.CRC,
                is_dir=zipinfo.is_dir(),
                is_file=not zipinfo.is_dir(),
            )
            for zipinfo in self._zipfile.filelist
        )

    def get_names(self) -> tuple[str, ...]:
        return tuple(self._zipfile.namelist())

    def print_tree(
        self,
        *,
        max_depth: int = 0,
        style: TreeStyle = "const",
    ) -> None:
        try:
            from bigtree.tree.construct import list_to_tree
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("The 'print_tree()' method requires the 'bigtree' dependency.")

        paths = [f"{self.file.name}/{member}" for member in self.get_names()]
        tree = list_to_tree(paths)  # type: ignore
        tree.show(max_depth=max_depth, style=style)

    def print_table(
        self,
        *,
        title: str | None = None,
        style: TableStyle = "markdown",
        sort_by: SortBy = "name",
        descending: bool = False,
        **kwargs: Any,
    ) -> None:
        try:
            from rich import box as RichBox
            from rich import print as richprint
            from rich.table import Table as RichTable
        except ModuleNotFoundError:  # pragma: no cover
            raise ModuleNotFoundError("The 'print_table()' method requires the 'rich' dependency.")

        if title is None:
            title = self.file.name

        members = self.get_members()
        table = RichTable(title=title, box=getattr(RichBox, style.upper()), **kwargs)
        table.add_column("Name", no_wrap=True)
        table.add_column("Date modified", no_wrap=True)
        table.add_column("Type", no_wrap=True)
        table.add_column("Size", no_wrap=True)
        table.add_column("Compressed Size", no_wrap=True)
        for member in sorted(members, key=lambda member: getattr(member, sort_by), reverse=descending):
            table.add_row(
                member.name,
                member.datetime.isoformat(),
                "File" if member.is_file else "Folder",
                member.size.human_readable(),
                member.compressed_size.human_readable(),
            )
        richprint(table)

    def extract(self, member: StrPath | ArchiveMember, *, destination: StrPath = Path.cwd()) -> Path:
        destination = realpath(destination)
        destination.mkdir(parents=True, exist_ok=True)

        name = get_member_name(member)
        self._zipfile.extract(member=name, path=destination, pwd=self._pwd)
        return destination / name

    def extractall(
        self, *, destination: StrPath = Path.cwd(), members: CollectionOf[StrPath | ArchiveMember] | None = None
    ) -> Path:
        destination = realpath(destination)
        destination.mkdir(parents=True, exist_ok=True)

        names: list[str] = []
        if members:
            for member in members:
                names.append(get_member_name(member))

        if names:
            self._zipfile.extractall(path=destination, members=names, pwd=self._pwd)
        else:
            self._zipfile.extractall(path=destination, pwd=self._pwd)

        return destination

    def read_bytes(self, member: StrPath | ArchiveMember) -> bytes:
        name = get_member_name(member)
        return self._zipfile.read(name, pwd=self._pwd)  # type: ignore

    def read_text(
        self,
        member: StrPath | ArchiveMember,
        *,
        encoding: str = "utf-8",
        errors: ErrorHandler = "strict",
    ) -> str:
        return self.read_bytes(member).decode(encoding, errors)

    def write(
        self,
        file: StrPath,
        *,
        arcname: StrPath | None = None,
    ) -> None:
        file = realpath(file)

        if arcname is None:
            arcname = file.name

        if not file.is_file():
            raise ValueError(f"The specified file '{file}' either does not exist or is not a regular file!")

        self._zipfile.write(file, arcname=arcname)

    def write_text(
        self,
        data: str,
        *,
        arcname: StrPath,
    ) -> None:
        self._zipfile.writestr(get_member_name(arcname), data)

    def write_bytes(
        self,
        data: bytes,
        *,
        arcname: StrPath,
    ) -> None:
        self._zipfile.writestr(get_member_name(arcname), data)

    def writeall(
        self,
        dir: StrPath,
        *,
        root: StrPath | None = None,
        glob: str = "*",
        recursive: bool = True,
    ) -> None:
        dir = realpath(dir)

        if not dir.is_dir():
            raise ValueError(f"The specified file '{dir}' either does not exist or is not a regular directory!")

        if root is None:
            root = dir.parent
        else:
            root = realpath(root)

        if not dir.is_relative_to(root):
            raise ValueError(f"{dir} must be relative to {root}")

        files = dir.rglob(glob) if recursive else dir.glob(glob)

        for file in files:
            arcname = file.relative_to(root)
            self.write(file, arcname=arcname)

    def close(self) -> None:
        self._zipfile.close()

    def __repr__(self) -> str:
        password = '"********"' if self.password else None
        return f'{self.__class__.__name__}("{self.file}", "{self.mode}", password={password})'



from __future__ import annotations

from datetime import datetime
from typing import Any

from pydantic import BaseModel, ByteSize, ConfigDict, ValidationInfo, field_validator

from archivefile._types import UTCDateTime


class ArchiveMember(BaseModel):
    """
    Represents a member of an archive file.
    This is immutable, hashable, and supports equality checking.
    """

    model_config = ConfigDict(frozen=True)

    name: str
    """Name of the archive member."""

    size: ByteSize = ByteSize(0)
    """Uncompressed size of the archive member."""

    compressed_size: ByteSize = ByteSize(0)
    """Compressed size of the archive member."""

    datetime: UTCDateTime = datetime.min
    """The time and date of the last modification to the archive member."""

    checksum: int = 0
    """CRC32 checksum if the archive is a ZipFile, RarFile, or SevenZipFile. Header checksum if archive is a TarFile."""

    is_dir: bool = False
    """True if the archive member is a directory, False otherwise."""

    is_file: bool = False
    """True if the archive member is a file, False otherwise."""

    @field_validator("*", mode="before")
    @classmethod
    def _use_default_value(cls, v: Any, info: ValidationInfo) -> Any:
        if v is None:
            return cls.model_fields[info.field_name].default  # type: ignore
        return v

    def __str__(self) -> str:
        """
        Simple string representation.
        This is equivalent to `ArchiveMember.name`
        """
        return self.name



from __future__ import annotations

from pathlib import Path
from tarfile import is_tarfile
from zipfile import is_zipfile

from py7zr import is_7zfile
from rarfile import is_rarfile, is_rarfile_sfx

from archivefile._models import ArchiveMember
from archivefile._types import StrPath


def realpath(path: StrPath) -> Path:
    """
    Get the real path of a given file or directory.

    Parameters
    ----------
    path : str or Path
        A string representing a path or a Path object.

    Returns
    -------
    Path
        The path after expanding the user's home directory and resolving any symbolic links.
    """
    return path.expanduser().resolve() if isinstance(path, Path) else Path(path).expanduser().resolve()


def is_archive(file: StrPath) -> bool:
    """
    Check whether the given archive file is a supported archive or not.

    Parameters
    ---------
    file : StrPath
        Path to the archive file.

    Returns
    -------
    bool
        True if the archive is supported, False otherwise.
    """
    file = realpath(file)

    if file.exists():
        return is_tarfile(file) or is_zipfile(file) or is_rarfile(file) or is_rarfile_sfx(file) or is_7zfile(file)
    else:
        return False


def get_member_name(member: StrPath | ArchiveMember) -> str:
    """Get the member name from a string, path, or ArchiveMember"""

    match member:
        case ArchiveMember():
            return member.name

        case Path():
            return member.relative_to(member.anchor).as_posix()

        case _:
            return member


def clamp_compression_level(level: int) -> int:
    """
    Pretty simple method to clamp compression level to a valid range
    """
    return max(0, min(level, 9))



from __future__ import annotations

from archivefile._core import ArchiveFile
from archivefile._enums import CompressionType
from archivefile._models import ArchiveMember
from archivefile._utils import is_archive
from archivefile._version import Version, _get_version

__version__ = _get_version()
__version_tuple__ = Version(*[int(i) for i in __version__.split(".")])

__all__ = [
    "ArchiveFile",
    "ArchiveMember",
    "is_archive",
    "CompressionType",
]



from __future__ import annotations

from datetime import datetime, timezone
from pathlib import Path
from typing import Annotated, Literal, TypeAlias, Union

from pydantic import AfterValidator
from typing_extensions import TypeVar

UTCDateTime = Annotated[datetime, AfterValidator(lambda dt: dt.astimezone(timezone.utc))]
"""Datetime that's always in UTC."""

StrPath: TypeAlias = Union[str, Path]

T = TypeVar("T")
CollectionOf: TypeAlias = Union[list[T], tuple[T, ...], set[T]]
"""Type alias representing a union of list, tuple, and set."""

OpenArchiveMode: TypeAlias = Literal[
    "r",
    "r:*",
    "r:",
    "r:gz",
    "r:bz2",
    "r:xz",
    "w",
    "w:",
    "w:gz",
    "w:bz2",
    "w:xz",
    "x",
    "x:",
    "x:",
    "x:gz",
    "x:bz2",
    "x:xz",
    "a",
    "a:",
]

TreeStyle: TypeAlias = Literal["ansi", "ascii", "const", "const_bold", "rounded", "double"]

TableStyle: TypeAlias = Literal[
    "ascii",
    "ascii2",
    "ascii_double_head",
    "square",
    "square_double_head",
    "minimal",
    "minimal_heavy_head",
    "minimal_double_head",
    "simple",
    "simple_head",
    "simple_heavy",
    "horizontals",
    "rounded",
    "heavy",
    "heavy_edge",
    "heavy_head",
    "double",
    "double_edge",
    "markdown",
]

SortBy: TypeAlias = Literal["name", "datetime", "size", "compressed_size", "checksum"]

ErrorHandler: TypeAlias = Literal[
    "strict", "ignore", "replace", "backslashreplace", "surrogateescape", "xmlcharrefreplace", "namereplace"
]

CompressionLevel: TypeAlias = Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, overload

from archivefile._adapters._base import BaseArchiveAdapter
from archivefile._models import ArchiveMember
from archivefile._utils import get_member_name, realpath
from rarfile import NoRarEntry, RarFile

if TYPE_CHECKING:
    from types import TracebackType

    from archivefile._enums import CompressionType
    from archivefile._types import (
        CollectionOf,
        CompressionLevel,
        ErrorHandler,
        OpenArchiveMode,
        SortBy,
        StrPath,
        TableStyle,
        TreeStyle,
    )
    from rarfile import RarInfo
    from typing_extensions import Generator, Self


class RarFileAdapter(BaseArchiveAdapter):
    # fmt: off
    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: OpenArchiveMode = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: CompressionLevel | None = None, **kwargs: Any) -> None: ...

    @overload
    def __init__(self, file: StrPath, mode: str = "r", *, password: str | None = None, compression_type: CompressionType | None = None, compression_level: int | None = None, **kwargs: Any) -> None: ...
    # fmt: on

    def __init__(
        self,
        file: StrPath,
        mode: OpenArchiveMode | str = "r",
        *,
        password: str | None = None,
        compression_type: CompressionType | None = None,
        compression_level: CompressionLevel | int | None = None,
        **kwargs: Any,
    ) -> None:
        self._file = realpath(file)
        self._mode = mode[0]
        self._password = password

        if self._mode == "r":
            self._rarfile = RarFile(self._file, mode=self._mode, **kwargs)
        else:
            raise NotImplementedError('Cannot write to a rar file. Rar files only support mode="r"!')

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        self._rarfile.close()

    @property
    def file(self) -> Path:
        return self._file

    @property
    def mode(self) -> OpenArchiveMode:
        return self._mode  # type: ignore

    @property
    def password(self) -> str | None:
        return self._password

    @property
    def compression_type(self) -> CompressionType | None:
        # RarFile doesn't support writing, so this will always be None
        return None

    @property
    def compression_level(self) -> CompressionLevel | None:
        # RarFile doesn't support writing, so this will always be None
        return None

    @property
    def adapter(self) -> str:
        return self.__class__.__name__

    def get_member(self, member: StrPath) -> ArchiveMember:
        name = get_member_name(member)

        try:
            # ZipFile and TarFile raise KeyError but RarFile raises it's own NoRarEntry
            # So for consistency's sake, we'll also raise KeyError here
            rarinfo: RarInfo = self._rarfile.getinfo(name)
        except NoRarEntry:
            raise KeyError(f"{name} not found in {self._file}")

        is_dir = True if rarinfo.filename.endswith("/") else False
        return ArchiveMember(
            name=rarinfo.filename,
            size=rarinfo.file_size,
            compressed_size=rarinfo.compress_size,
            datetime=datetime(*rarinfo.date_time),
            checksum=rarinfo.CRC,
            is_dir=is_dir,
            is_file=not is_dir,
        )

    def get_members(self) -> Generator[ArchiveMember]:
        return (
            ArchiveMember(
                name=rarinfo.filename,
                size=rarinfo.file_size,
                compressed_size=rarinfo.compress_size,
                datetime=datetime(*rarinfo.date_time),
                checksum=rarinfo.CRC,
                is_dir=True if rarinfo.filename.endswith("/") else False,
                is_file=False if rarinfo.filename.endswith("/") else True,
            )
            for rarinfo in self._rarfile.infolist()
        )

    def get_names(self) -> tuple[str, ...]:
        return tuple(self._rarfile.namelist())

    def print_tree(
        self,
        *,
        max_depth: int = 0,
        style: TreeStyle = "const",
    ) -> None:
        try:
            from bigtree.tree.construct import list_to_tree
        except ModuleNotFoundError:
            raise ModuleNotFoundError("The 'print_tree()' method requires the 'bigtree' dependency.")

        paths = [f"{self.file.name}/{member}" for member in self.get_names()]
        tree = list_to_tree(paths)  # type: ignore
        tree.show(max_depth=max_depth, style=style)

    def print_table(
        self,
        *,
        title: str | None = None,
        style: TableStyle = "markdown",
        sort_by: SortBy = "name",
        descending: bool = False,
        **kwargs: Any,
    ) -> None:
        try:
            from rich import box as RichBox
            from rich import print as richprint
            from rich.table import Table as RichTable
        except ModuleNotFoundError:
            raise ModuleNotFoundError("The 'print_table()' method requires the 'rich' dependency.")

        if title is None:
            title = self.file.name

        members = self.get_members()
        table = RichTable(title=title, box=getattr(RichBox, style.upper()), **kwargs)
        table.add_column("Name", no_wrap=True)
        table.add_column("Date modified", no_wrap=True)
        table.add_column("Type", no_wrap=True)
        table.add_column("Size", no_wrap=True)
        table.add_column("Compressed Size", no_wrap=True)
        for member in sorted(members, key=lambda member: getattr(member, sort_by), reverse=descending):
            table.add_row(
                member.name,
                member.datetime.isoformat(),
                "File" if member.is_file else "Folder",
                member.size.human_readable(),
                member.compressed_size.human_readable(),
            )
        richprint(table)

    def extract(self, member: StrPath | ArchiveMember, *, destination: StrPath = Path.cwd()) -> Path:
        destination = realpath(destination)
        destination.mkdir(parents=True, exist_ok=True)

        name = get_member_name(member)

        try:
            # ZipFile and TarFile raise KeyError but RarFile raises it's own NoRarEntry
            # So for consistency's sake, we'll also raise KeyError here
            self._rarfile.extract(member=name, path=destination, pwd=self._password)
        except NoRarEntry:
            raise KeyError(f"{name} not found in {self._file}")

        return destination / name

    def extractall(
        self, *, destination: StrPath = Path.cwd(), members: CollectionOf[StrPath | ArchiveMember] | None = None
    ) -> Path:
        destination = realpath(destination)
        destination.mkdir(parents=True, exist_ok=True)

        names: set[str] = set()
        if members:
            all_members = self._rarfile.namelist()
            for member in members:
                name = get_member_name(member)
                if name in all_members:
                    names.add(name)
                else:
                    raise KeyError(f"{name} not found in {self._file}")

        self._rarfile.extractall(path=destination, members=names, pwd=self._password)
        return destination

    def read_bytes(self, member: StrPath | ArchiveMember) -> bytes:
        name = get_member_name(member)

        if name.endswith("/"):
            return b""

        try:
            # ZipFile and TarFile raise KeyError but RarFile raises it's own NoRarEntry
            # So for consistency's sake, we'll also raise KeyError here
            return self._rarfile.read(name, pwd=self._password)
        except NoRarEntry:
            raise KeyError(f"{name} not found in {self._file}")

    def read_text(
        self,
        member: StrPath | ArchiveMember,
        *,
        encoding: str = "utf-8",
        errors: ErrorHandler = "strict",
    ) -> str:
        return self.read_bytes(member).decode(encoding, errors)

    def write(
        self,
        file: StrPath,
        *,
        arcname: StrPath | None = None,
    ) -> None:
        raise NotImplementedError('Cannot write to a rar file. Rar files only support mode="r"!')

    def write_text(
        self,
        data: str,
        *,
        arcname: StrPath,
    ) -> None:
        raise NotImplementedError('Cannot write to a rar file. Rar files only support mode="r"!')

    def write_bytes(
        self,
        data: bytes,
        *,
        arcname: StrPath,
    ) -> None:
        raise NotImplementedError('Cannot write to a rar file. Rar files only support mode="r"!')

    def writeall(
        self,
        dir: StrPath,
        *,
        root: StrPath | None = None,
        glob: str = "*",
        recursive: bool = True,
    ) -> None:
        raise NotImplementedError('Cannot write to a rar file. Rar files only support mode="r"!')

    def close(self) -> None:
        self._rarfile.close()

    def __repr__(self) -> str:
        password = '"********"' if self.password else None
        return f'{self.__class__.__name__}("{self.file}", "{self.mode}", password={password})'
